/*
  This program is passed a filename which contains Performer-loadable data.
  The file is loaded using the appropriate Performer file loader, and a
  description of the scenegraph is printed to stdout.

  This program was written to help debug the output of the pfdBuilder and
  pfdBreakup utilities.  It is by no means complete.  It does not handle
  indexed data, and non-geometric data such as pfVolume nodes.  It is useful
  to see details about what you're actually looking at, as the pfdBuilder and
  pdfBreakup might substantially change the model.

 */

#include <stdio.h>
#include <Performer/pf.h>
#include <Performer/pf/pfNode.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pfdu.h>

static void _printScenegraph(FILE *fp, pfNode *node, int scoot) ;
static void _printGeoSet(FILE *fp, pfGeoSet *geoset, int scoot) ;

int main (int argc, char **argv) {

  if (argc!=2) {
    fprintf(stderr,"Usage: %s filename\n",argv[0]) ;
    return 1 ;
  }

  printf("Printing tree for file \"%s\"\n",argv[1]) ;

  pfInit();
  
  pfMultiprocess(0); 

  pfConfig();

  char path[PF_MAXSTRING] ;
  if (pfGetFilePath()) {
    strcpy(path,pfGetFilePath()) ;
    strcat(path,":.:/usr/share/Performer/data") ;
  } else {
    strcpy(path,":.:/usr/share/Performer/data") ;
  }
  pfFilePath(path);

  pfNode *node = pfdLoadFile(argv[1]) ;

  if (node)
    _printScenegraph(stdout, node, 0) ;

  pfExit() ;

}

static void _printScenegraph(FILE *fp, pfNode *node, int scoot) {

  const char *primType[] = { "POINTS",  "LINES", "LINESTRIPS",  "TRIS",  "QUADS",  "TRISTRIPS",  "FLAT_LINESTRIPS",  "FLAT_TRISTRIPS",  "POLYS",  "TRIFANS",  "FLAT_TRIFANS" } ;

  int i ;
  for (i=0; i<scoot; i++)
    fprintf(fp, " ") ;
  fprintf(fp, "%s: \"%s\"",node->getTypeName(), node->getName()) ;
  if (node->isOfType(pfGroup::getClassType())) {
    pfGroup *group ;
    group = (pfGroup*)node ;
    fprintf(fp, " %d children\n",group->getNumChildren()) ;
    for (i=0; i<group->getNumChildren(); i++) {
      _printScenegraph(fp, group->getChild(i), scoot+1) ;
    }
  } else if (node->isOfType(pfGeode::getClassType())) { 
    pfGeode *geode ;
    geode = (pfGeode*)node ;
    fprintf(fp, " with %d geoSets\n",
	    geode->getNumGSets()) ;
    for (i=0; i<geode->getNumGSets(); i++) {
      fprintf(fp, "%*sgeoSet %d, %s, contains %d primitives\n",
	      scoot+1," ",i+1,primType[geode->getGSet(i)->getPrimType()],
	      geode->getGSet(i)->getNumPrims()) ;
      _printGeoSet(fp, geode->getGSet(i), scoot+3) ;
    }
  } 
}

static void _printGeoSet(FILE *fp, pfGeoSet *geoset, int scoot) {
  pfVec3 *coords ;
  pfVec3 *norms;
  pfVec4 *colors;
  ushort  *icoords ;
  ushort  *inorms ;
  ushort  *icolors ;
  int normBind ;
  int colorBind ;
  int inorm ;
  int icolor ;

#if 0
  printf("\n\n\n") ;
  geoset->print(1, 0, "foo", stdout) ;
  printf("\n\n\n") ;
  fflush(stdout) ;
#endif

  geoset->getAttrLists(PFGS_COORD3, (void **)&coords, &icoords);
  fprintf(fp, "%*s%d pfVec3 coords\n",scoot," ",
	  (int)(pfGetSize(coords)/(3*sizeof(float)))) ;

  geoset->getAttrLists(PFGS_NORMAL3, (void **)&norms, &inorms);
  if (norms)
    fprintf(fp, "%*s%d pfVec3 norms\n",scoot," ",
	    (int)(pfGetSize(norms)/(3*sizeof(float)))) ;
  else
    fprintf(fp, "%*sno norms\n",scoot," ") ;

  geoset->getAttrLists(PFGS_COLOR4, (void **)&colors, &icolors);
  if (colors)
    fprintf(fp, "%*s%d pfVec4 colors\n",scoot," ",
	    (int)(pfGetSize(colors)/(4*sizeof(float)))) ;
  else
    fprintf(fp, "%*sno colors\n",scoot," ") ;

  normBind = geoset->getAttrBind(PFGS_NORMAL3) ;
  if (icoords)
    fprintf(fp,"%*sgeoSet is indexed\n",scoot," ") ;
  else 
    fprintf(fp,"%*sgeoSet is not indexed\n",scoot," ") ;
      
  if (normBind == PFGS_OFF) {
    fprintf(fp,"%*snormals are OFF\n",scoot+1," ") ;
  } else if (normBind == PFGS_OVERALL) {
    if (inorms) {
      inorm = inorms[0] ;
    } else {
      inorm = 0 ;
    }
    fprintf(fp,"%*snormals are OVERALL: %f %f %f\n",scoot+1," ",
	    norms[inorm][0], norms[inorm][1], 
	    norms[inorm][2]) ;
  }

  colorBind = geoset->getAttrBind(PFGS_COLOR4) ;
  if (colorBind == PFGS_OFF) {
    fprintf(fp,"%*scolors are OFF\n",scoot+1," ") ;
  } else if (colorBind == PFGS_OVERALL) {
    if (icolors) {
      icolor = icolors[0] ;
    } else {
      icolor = 0 ;
    }
    fprintf(fp,"%*scolors are OVERALL: %f %f %f %f\n",scoot," ",
	    colors[icolor][0], colors[icolor][1], 
	    colors[icolor][2], colors[icolor][3]) ;
  }

  int i, j ;

  // offset to start of a primitive (can have more than one vertice)
  int p = 0 ;

  int primType = geoset->getPrimType() ;
  for (i=0; i<geoset->getNumPrims(); i++) {
    int primLen = geoset->getPrimLength(i) ;
    fprintf(fp,"%*sprimitive %d prim length is %d\n",
	    scoot," ",i+1,primLen) ;

    if (normBind == PFGS_PER_PRIM) {
      if (inorms) {
	inorm = inorms[i] ;
      } else {
	inorm = i ;
      }
      fprintf(fp,"%*snormals are PER_PRIM: %f %f %f\n",scoot+1," ",
	      norms[inorm][0], norms[inorm][1], 
	      norms[inorm][2]) ;
    }

    if (colorBind == PFGS_PER_PRIM) {
      if (icolors) {
	icolor = icolors[i] ;
      } else {
	icolor = i ;
      }
      fprintf(fp,"%*scolors are PER_PRIM: %f %f %f %f\n",scoot+1," ",
	      colors[icolor][0], colors[icolor][1], 
	      colors[icolor][2], colors[icolor][3]) ;
    }

    int numVerts ;
    // for triangles, the primitive length is number of triangles
    if (primType == PFGS_TRIS)
      numVerts = primLen*3 ;
    // for quads, the primitive length is number of quads
    else if (primType == PFGS_QUADS) 
      numVerts = primLen*4 ;
    else // for others, prim length = number of verts
      numVerts = primLen ;
    fprintf(fp,"%*s%d vertices:\n",
	    scoot+1," ",numVerts) ;

    for (j=0; j<numVerts; j++) {
      int icoord ;
      if (icoords)
	icoord = icoords[p] ;
      else
	icoord = p ;
      fprintf(fp,"%*s%3.3d: coords: %f %f %f\n",
	      scoot+1," ",j+1,
	      coords[icoord][0], coords[icoord][1], coords[icoord][2]) ;
      
      if (normBind == PFGS_PER_VERTEX) {
	if (inorms) {
	  inorm = inorms[p] ;
	} else {
	  inorm = p ;
	}
	fprintf(fp,"%*snormals: %f %f %f\n",scoot+5," ",
		norms[inorm][0], norms[inorm][1], 
		norms[inorm][2]) ;
      }

      if (colorBind == PFGS_PER_VERTEX) {
	if (icolors) {
	  icolor = icolors[p] ;
	} else {
	  icolor = p ;
	}
	fprintf(fp,"%*scolors: %f %f %f %f\n",scoot+6," ",
		colors[icolor][0], colors[icolor][1], 
		colors[icolor][2], colors[icolor][3]) ;
      }
      p++ ;
    }
  }
  //printf("!!!!!!!!p=%d\n",p) ;
}
